# coding: utf-8
# @author Jonathan Raiman


# Vectors with named dimensions. Infinite dimensional objects with simple vector calculus operations defined.
class NamedVector 
	attr_reader :keys
	alias :dimensions :keys

	# creates a new named vector
	# @param dims [Hash<String,Fixnum>, Hash<Symbol,Fixnum>, Array<String>, Array<Symbol>] the dimensions to initialize the vectors with, either as a hash of assignments or simply the names of the dimensions.
	def initialize(*dims)
		@vector = Hash.new
		@keys   = Set.new
		import_dimensions(dims)
	end

	# imports the dimension information in Array form
	# @param dims [Array<Array>,Array<String>, Array<Symbol>, Array<Hash>, Array<Fixnum>, Array<Array>] values to initialize
	def import_dimensions(dims)
		if dims && dims.first
			case dims.first
			when Array
				array = dims.first
				if array.first.class == String
					@keys = Set.new(array.map {|i| i.to_s})
					array.each {|d| self[d]+=1}
				end
			when Hash
				hash = dims.first
				hash.keys.each {|d| modify_dimension(d, hash[d])}
			when String, Symbol, Fixnum
				@keys = Set.new(dims.map {|i| i.to_s})
				dims.each {|d| new_dimension(d)}
			end
		else
			@keys = []
		end
	end

	# Perform a dot product with another vector
	# @param value [Hash, Fixnum, Float, NamedVector] the vector or scalar to multiply this vector with. Scalar multiplication is destructive.
	def *(value)
		case value
		when Fixnum, Float  #scalar multiplication product
			each_dimension {|i| @vector[i] = self[i]*value}
		when NamedVector #dot product
			NamedVector.dot_product(self, value)
		when Hash #dot product
			NamedVector.add_dimension_iterator(value)
			NamedVector.dot_product(self,value)
		else
			raise TypeError, "#{value.class} cannot be coerced to NamedVector."
		end
	end

	# adds a dimension iterator (iterate through the keys) to a Hash or NamedVector
	# @param obj [Hash, NamedVector] the object to add the method to.
	def self.add_dimension_iterator(obj)
		obj.define_singleton_method(:each_dimension) do |&block|
			self.keys.each do |key|
				block.call(key)
			end
		end
	end

	# Perform a dot product with another vector
	# @param value [Hash, NamedVector] the vector to multiply this vector with.
	def dot_product(value)
		NamedVector.dot_product(self, value)
	end

	# Perform a dot product with another vector
	# @param a [Hash, NamedVector] the vector or scalar to multiply this vector with.
	# @param b [Hash, NamedVector] the vector or scalar to multiply this vector with.
	def self.dot_product(a, b)
		sum    = 0.0
		if a.keys.length < b.keys.length
			a.each_dimension {|i| sum+=a[i]*b[i]}
		else
			b.each_dimension {|i| sum+=b[i]*a[i]}
		end
		sum
	end

	# iterates through each dimension of a vector
	def each_dimension
		@keys.each do |key|
			yield(key)
		end
	end

	# fetches the value at the dimension specified
	# @param key [Symbol, String] the dimension requested
	def [](key)
		@vector.fetch(key.to_s,0.0)
	end

	def modify_dimension(key,value)
		new_dimension(key.to_s)
		@vector[key.to_s] = value
	end

	# assigns a value at the dimension specified
	# @param key [Symbol, String] the dimension requested
	# @param value [Fixnum, Float] the assignment for this dimension
	def []=(key,value)
		modify_dimension(key,value)
	end

	def method_missing(mid, *args) # :nodoc:
		mname = mid.id2name
		len = args.length
		if mname.chomp!('=')
			if len != 1
				raise ArgumentError, "wrong number of arguments (#{len} for 1)", caller(1)
			end
			modify_dimension(mname, args[0])
		elsif len == 0
			@vector.fetch(mname,0.0)
		else
			raise NoMethodError, "undefined method `#{mid}' for #{self}", caller(1)
		end
	end

	def new_dimension(name)
		sym_name = name.to_sym
		unless respond_to?(sym_name)
			@keys << name
			define_singleton_method(sym_name) { @vector.fetch(name,0.0) }
			define_singleton_method("#{sym_name}=") { |x| @vector[name] = x }
		end
		name
	end
	protected :new_dimension

	# the squared Euclidean norm of a vector
	def squared_norm
		norm    = 0.0
		each_dimension {|i| norm+=self[i]**2}
		norm
	end

	def eql?(other)
		false unless other.keys == @keys
		all_keys = other.keys.merge(@keys)
		all_keys.each do |key|
			if other[key] != self[key] then return false end
		end
		true
	end

	def to_s
		"<#NamedVector @keys=\"#{@keys.inspect}\", @vector=\"#{@vector.inspect}\">"
	end

	def to_a
		@vector.to_a
	end

	alias :to_ary :to_a

	alias :== :eql?

	# the Euclidean norm of a vector
	def norm
		Math.sqrt(squared_norm)
	end

	# normalizes the vector destructively.
	def normalize
		current_norm = norm
		if current_norm > 0 then self*(1.0/current_norm) end
		self
	end

	# Mongo retrieval
	# @param doc [BSON::OrderedHash] the mongo document to retrieve the vector from.
	def self.to_mongo(hash)
		if hash.is_a?(Hash) then hash
		else
			{"keys"=> hash.keys.to_a, "dimensions" => hash.keys.map {|i| hash[i]}}
		end
	end

	# Mongo conversion
	# @param doc [BSON::OrderedHash] the mongo document to retrieve the vector from.
	def self.from_mongo(doc)
		if doc.is_a?(self) then doc
		else
			hash = {}
			doc["keys"].each_with_index do |k, index|
				hash[k] = doc["dimensions"][index]
			end
			NamedVector.new(hash)
		end
	end

	private :new_dimension
	private :method_missing
	private :modify_dimension

end